//
//  UpdaterController.m
//  Myth 1.4 Updater OSX
//
//  Created by Chris Dreessen on Wed May 28 2003.
//  Copyright (c) 2003 Chris Dreessen. All rights reserved.
//

#import "UpdaterController.h"
#import <CoreFoundation/CoreFoundation.h>
#import <sys/types.h>
#import <unistd.h>
#import <pwd.h>
#import <errno.h>

enum {
	kRWXUserAccessMask  = 0x01C0,
	kReadAccessUser   = 0x0100,
	kWriteAccessUser  = 0x0080,
	kExecuteAccessUser  = 0x0040,
	
	kRWXGroupAccessMask  = 0x0038,
	kReadAccessGroup  = 0x0020,
	kWriteAccessGroup  = 0x0010,
	kExecuteAccessGroup  = 0x0008,
	
	kRWXOtherAccessMask  = 0x0007,
	kReadAccessOther  = 0x0004,
	kWriteAccessOther  = 0x0002,
	kExecuteAccessOther  = 0x000
};

const CFStringRef kStringMythIIVersion14Name = (CFStringRef)@"Myth II";
const CFStringRef kStringMythIIOldName = (CFStringRef)@"Myth II (Old)";
const CFStringRef kStringMythIIEELName = (CFStringRef)@"Myth 2 EEL";
const CFStringRef kStringMythIIPluginsFolderName = (CFStringRef)@"plugins";
const CFStringRef kStringDataName = (CFStringRef)@"Installer Data";
const CFStringRef kStringOldMythIIFiles = (CFStringRef)@"Myth II Files (old)";

const OSType kMythIICreatorCode = 'M2SB';
const OSType kMythIITypeCode = 'APPL';

const NSTimeInterval kRunloopUpdateDuration = 0.1;

enum {
	kNumberOfCopyBufferPages = 128
};

enum {
	cdApplicationFatalError = -307/*smResrvErr*/
};

OSErr FSCopyObject( const FSRef* ref, const FSRef* destDirectory, FSRef* newRef, const HFSUniStr255* new_name)
{
	FSRef* destFile;
	OSErr err = noErr;
	
	if (newRef != NULL)
		destFile = newRef;
	else
		destFile = alloca(sizeof(FSRef));
	
	if (!destFile)
		err = memFullErr;
	else
	{
		FSCatalogInfo info;
		HFSUniStr255 name;
		
		err = FSGetCatalogInfo(ref, kFSCatInfoSettableInfo, &info, &name, NULL, NULL);
		
		if (err == noErr)
		{
			if (new_name != NULL)
				err = FSCreateFileUnicode( destDirectory, new_name->length, new_name->unicode, kFSCatInfoNone, NULL, destFile, NULL);
			else
				err = FSCreateFileUnicode( destDirectory, name.length, name.unicode, kFSCatInfoNone, NULL, destFile, NULL);

			if (err == noErr)
			{
				void* buffer;
				size_t buffer_size = getpagesize()*kNumberOfCopyBufferPages;
				
				buffer = malloc(buffer_size);
				
				if (buffer == NULL)
					err = memFullErr;
				else
				{
					CatPositionRec forkIterator;
					UInt64 forkSize;
					ByteCount bytesBuffered, bytesCopied;
					SInt16 readRefNum, writeRefNum;
					
					forkIterator.initialize = 0;
					
					while ( (err == noErr) && (noErr == (err = FSIterateForks(ref, &forkIterator, &name, &forkSize, NULL))) )
					{
						if (forkSize > 0)
						{
							err = FSOpenFork( ref, name.length, name.unicode, fsRdPerm, &readRefNum);
							
							if (err == noErr)
							{
								err = FSOpenFork( destFile, name.length, name.unicode, fsWrPerm, &writeRefNum);
								
								if (err == errFSForkNotFound)
								{
									err = FSCreateFork( destFile, name.length, name.unicode );
									if (err == noErr)
										err = FSOpenFork( destFile, name.length, name.unicode, fsWrPerm, &writeRefNum);
								}
								
								if (err == noErr)
								{
									err = FSAllocateFork( writeRefNum, kFSAllocAllOrNothingMask, fsFromStart, 0, forkSize, NULL);
									
									if (err == noErr)
									{
										bytesCopied = 0;
										
										while (bytesCopied < forkSize && err == noErr)
										{
											err = FSReadFork( readRefNum, fsFromStart, bytesCopied, buffer_size, buffer, &bytesBuffered);
		
											if (err == noErr || err == eofErr)
											{
												err = FSWriteFork( writeRefNum, fsFromStart, bytesCopied, bytesBuffered, buffer, &bytesBuffered);
												bytesCopied+=bytesBuffered;
											}
										}
									}
									
									if (err == noErr)
										err = FSFlushFork(writeRefNum);
										
									FSCloseFork(writeRefNum);
								}
								
								FSCloseFork(readRefNum);
							}
						}
					}
	
					if (err == errFSNoMoreItems)
						err = noErr;
						
					free(buffer);
				}
				
				if (err != noErr)
					FSDeleteObject(destFile);
			}
		}
		
		if (err == noErr)
			err = FSSetCatalogInfo( destFile, kFSCatInfoSettableInfo, &info);
	}
	
	return err;
}

OSErr CDUnusedNamePermutation(const FSRef* container, HFSUniStr255* name)
{
	OSErr err = noErr;
	CFStringRef originalName = CFStringCreateWithCharacters(kCFAllocatorDefault, name->unicode, name->length);
	CFMutableStringRef safeName = NULL;
	FSRef attemptedNewObject;
	int iteration = 1;
	
	while (noErr == FSMakeFSRefUnicode( container, name->length, name->unicode, kTextEncodingUnknown, &attemptedNewObject))
	{
		if (safeName)
			CFRelease(safeName);
		
		safeName = CFStringCreateMutableWithExternalCharactersNoCopy(kCFAllocatorDefault, name->unicode, name->length, 255, kCFAllocatorNull);
		CFStringReplaceAll(safeName, originalName);
		
		iteration++;
		CFStringAppendFormat(safeName, NULL, CFSTR(" %d"), iteration);
		name->length = CFStringGetLength(safeName);
	}
	
	CFRelease(originalName);
	
	if (safeName != NULL)
		CFRelease(safeName);
		
	return err;
}

OSErr CDMoveObject( const FSRef* ref, const FSRef* destDirectory, FSRef* newRef)
{
	OSErr err;
	
	err = FSMoveObject(ref, destDirectory, newRef);
	
	if (err != noErr)
	{
		err = FSCopyObject(ref, destDirectory, newRef, NULL);
		if (err == noErr)
		{
			err = FSDeleteObject(ref);
		}
	}
	
	return err;
}

@implementation UpdaterController

-(void) awakeFromNib
{
	[progBar setDisplayedWhenStopped:NO];
}

-(void) chooseLocation:(id) sender
{
	NSOpenPanel* op = [NSOpenPanel openPanel];
	[op setAllowsMultipleSelection:NO];
	[op setCanChooseDirectories:YES];
	[op setCanChooseFiles:NO];
	
	[op beginSheetForDirectory:nil file:nil types:nil modalForWindow:window modalDelegate:self didEndSelector:@selector(openPanelDidEnd:returnCode:contextInfo:) contextInfo:nil];
}

-(void) openPanelDidEnd:(NSOpenPanel *)sheet returnCode:(int)returnCode contextInfo:(void *)contextInfo
{
	CFURLRef url = NULL;
	NSArray* files = [sheet URLs];
	if ([files count] > 0)
		url = (CFURLRef)[files objectAtIndex:0];
	BOOL success = NO;
	
	if (url)
		if (CFURLGetFSRef(url, &installLocation))
		{
			FSCatalogInfo info;
			if (noErr == FSGetCatalogInfo(&installLocation, kFSCatInfoPermissions | kFSCatInfoNodeFlags, &info, NULL, NULL, NULL))
			{
				FSPermissionInfo* permissions = (FSPermissionInfo*)(info.permissions);
				
				if ( ((info.nodeFlags & kFSNodeLockedMask) == 0) && ( ((permissions->mode & (kReadAccessOther | kWriteAccessOther)) != 0) ||
						((getgid() == permissions->groupID || getegid() == permissions->groupID) && ((permissions->mode & (kReadAccessGroup | kWriteAccessGroup)) != 0)) ||
						((getuid() == permissions->userID || geteuid() == permissions->userID) && ((permissions->mode & (kReadAccessUser | kWriteAccessUser)) != 0)) ) )
				{
					success = YES;
					[pathField setStringValue:[(NSURL*)url path]];
				}
			}
		}
	
	[installButton setEnabled:success];
}

-(void) install:(id) sender
{
	FSRef destinationPluginsFolder, sourcePluginsFolder, workingFileA, workingFileB, workingFileC, sourceFile, sourceFolder, trashFolder;
	FSIterator iterator;
	double progress = 0.0;
	HFSUniStr255 name;
	OSErr err = noErr;
	CFURLRef resourcesURL = CFBundleCopyResourcesDirectoryURL(CFBundleGetMainBundle());
	CFURLRef dataURL = CFURLCreateCopyAppendingPathComponent(kCFAllocatorDefault, resourcesURL, kStringDataName, YES);
	
	if (!CFURLGetFSRef(dataURL, &sourceFolder))
	{
		err = fnfErr;
		/*Installer damaged*/
		err = cdApplicationFatalError;
	}
	else
	{
		err = FSFindFolder( kUserDomain, kTrashFolderType, YES, &trashFolder);
		
		if (err == noErr)
		{
			name.length = CFStringGetLength(kStringOldMythIIFiles);
			CFStringGetCharacters(kStringOldMythIIFiles, CFRangeMake(0, name.length), name.unicode);
			
			err = CDUnusedNamePermutation(&trashFolder, &name);

			if (err == noErr)
				err = FSCreateDirectoryUnicode(&trashFolder, name.length, name.unicode, kFSCatInfoNone, NULL, &trashFolder, NULL, NULL);
		}
		
		if (err == noErr)
		{
			name.length = CFStringGetLength(kStringMythIIPluginsFolderName);
			CFStringGetCharacters(kStringMythIIPluginsFolderName, CFRangeMake(0, name.length), name.unicode);
			
			err = FSMakeFSRefUnicode( &sourceFolder, name.length, name.unicode, kTextEncodingUnknown, &sourcePluginsFolder);
			if (err == noErr)
			{
				FSCatalogInfo info;
				err = FSGetCatalogInfo(&sourcePluginsFolder, kFSCatInfoValence, &info, NULL, NULL, NULL);
				
				if (err == noErr)
				{
					[progBar setMaxValue:(double)(info.valence + 2)];
					[progBar setUsesThreadedAnimation:YES];
					[progBar setDoubleValue:progress];
					[progBar startAnimation:self];
					[pathField setEnabled:NO];
					[installButton setEnabled:YES];
					[locationButton setEnabled:NO];
					
					[[NSRunLoop currentRunLoop] runUntilDate:[NSDate dateWithTimeIntervalSinceNow:kRunloopUpdateDuration]];
					
/* Application */
					name.length = CFStringGetLength(kStringMythIIVersion14Name);
					CFStringGetCharacters(kStringMythIIVersion14Name, CFRangeMake(0, name.length), name.unicode);
					
					err = FSMakeFSRefUnicode( &sourceFolder, name.length, name.unicode, kTextEncodingUnknown, &sourceFile);
					
					if (err == noErr)
					{
						err = FSOpenIterator( &installLocation, kFSIterateFlat, &iterator);
						if (err == noErr)
						{
							ItemCount returnedItemCount;
							FInfo* finderInfo;
							
							while ((err == noErr) && (noErr == (err = FSGetCatalogInfoBulk(iterator, 1, &returnedItemCount, NULL, kFSCatInfoFinderInfo, &info, &workingFileA, NULL, NULL))))
							{
								finderInfo = (FInfo*)(info.finderInfo);
								
								if (finderInfo->fdType == kMythIITypeCode && finderInfo->fdCreator == kMythIICreatorCode)
								{
									err = FSMakeFSRefUnicode( &installLocation, name.length, name.unicode, kTextEncodingUnknown, &workingFileB);
									if (err == noErr)
									{
										if (noErr != FSCompareFSRefs(&workingFileA, &workingFileB))
										{
											err = CDMoveObject( &workingFileB, &trashFolder, NULL);
										}
										
										err = FSRenameUnicode(&workingFileA, name.length, name.unicode, kTextEncodingUnknown, &workingFileA);
									}
									else
										err = noErr;
									
									break;
								}
							}
							
							if (err == noErr)
							{
								name.length = CFStringGetLength(kStringMythIIOldName);
								CFStringGetCharacters(kStringMythIIOldName, CFRangeMake(0, name.length), name.unicode);
								err = FSMakeFSRefUnicode( &installLocation, name.length, name.unicode, kTextEncodingUnknown, &workingFileC);
								if (err == noErr)
									err = CDMoveObject( &workingFileC, &trashFolder, NULL);
								else
									err = noErr;

								err = FSCopyObject(&sourceFile, &installLocation, &workingFileB, &name);
								FSExchangeObjects(&workingFileA, &workingFileB);
							}
							else if (err == errFSNoMoreItems)
							{
								err = FSMakeFSRefUnicode( &installLocation, name.length, name.unicode, kTextEncodingUnknown, &workingFileB);
								if (err == noErr)
									err = CDMoveObject( &workingFileB, &trashFolder, NULL);
								else
									err = noErr;
								if (err == noErr)
									err = FSCopyObject(&sourceFile, &installLocation, &workingFileB, &name);
							}
							
							if (err == noErr)
								err = FSCloseIterator(iterator);
							else
								(void)FSCloseIterator(iterator);
							
							if (err == noErr)
							{
								progress+=1.0;
								[progBar setDoubleValue:progress];
								[[NSRunLoop currentRunLoop] runUntilDate:[NSDate dateWithTimeIntervalSinceNow:kRunloopUpdateDuration]];

							}
						}
					}
					else
					{
						/*Damaged Installer*/
						err = cdApplicationFatalError;
					}
	/* Copy EEL */
					if (err == noErr)
					{
						name.length = CFStringGetLength(kStringMythIIEELName);
						CFStringGetCharacters(kStringMythIIEELName, CFRangeMake(0, name.length), name.unicode);
						
						err = FSMakeFSRefUnicode( &sourceFolder, name.length, name.unicode, kTextEncodingUnknown, &sourceFile);
						if (err == noErr)
						{
							err = FSMakeFSRefUnicode( &installLocation, name.length, name.unicode, kTextEncodingUnknown, &workingFileA);
							if (err == noErr)
							{
								err = CDMoveObject( &workingFileA, &trashFolder, NULL);
							}
							else
								err = noErr;
							
							if (err == noErr)
								err = FSCopyObject(&sourceFile, &installLocation, NULL, NULL);
								
							if (err == noErr)
							{
								progress+=1.0;
								[progBar setDoubleValue:progress];
								[[NSRunLoop currentRunLoop] runUntilDate:[NSDate dateWithTimeIntervalSinceNow:kRunloopUpdateDuration]];

							}
						}
						else
						{
							/*Damaged Installer*/
							err = cdApplicationFatalError;
						}
					}
/* Copy plugins */
					if (err == noErr)
					{
						name.length = CFStringGetLength(kStringMythIIPluginsFolderName);
						CFStringGetCharacters(kStringMythIIPluginsFolderName, CFRangeMake(0, name.length), name.unicode);
						
						err = FSMakeFSRefUnicode( &installLocation, name.length, name.unicode, kTextEncodingUnknown, &destinationPluginsFolder);
						
						if (err != noErr)
							err = FSCreateDirectoryUnicode(&installLocation, name.length, name.unicode, kFSCatInfoNone, NULL, &destinationPluginsFolder, NULL, NULL);
						
						if (err == noErr)
						{
							err = FSOpenIterator( &sourcePluginsFolder, kFSIterateFlat, &iterator);
							if (err == noErr)
							{
								ItemCount returnedItemCount;
								
								while ((err == noErr) && (noErr == (err = FSGetCatalogInfoBulk(iterator, 1, &returnedItemCount, NULL, kFSCatInfoNone, NULL, &sourceFile, NULL, &name))))
								{
									err = FSMakeFSRefUnicode( &destinationPluginsFolder, name.length, name.unicode, kTextEncodingUnknown, &workingFileA);
									if (err == noErr)
									{
										err = CDMoveObject( &workingFileA, &trashFolder, NULL);
									}
									else
										err = noErr;
									
									if (err == noErr)
										err = FSCopyObject(&sourceFile, &destinationPluginsFolder, NULL, NULL);
										
									if (err == noErr)
									{
										progress+=1.0;
										[progBar setDoubleValue:progress];
										[[NSRunLoop currentRunLoop] runUntilDate:[NSDate dateWithTimeIntervalSinceNow:kRunloopUpdateDuration]];

									}
								}
								
								if (err == errFSNoMoreItems)
									err = FSCloseIterator(iterator);
								else
									(void)FSCloseIterator(iterator);
							}
						}
					}
					/*Done installation*/
				}
				
			}
			else
			{
				/*Installer Damaged*/
				err = cdApplicationFatalError;
			}
		}
		else
		{
			/*Trash not accessible*/
			NSRunAlertPanel(nil, @"The trash could not be accessed.", @"Quit", nil, nil);
		}
	}
	
	CFRelease(dataURL);
	CFRelease(resourcesURL);
	
	if (err == cdApplicationFatalError)
	{
		NSRunCriticalAlertPanel(nil, @"The installer is corrupted, and cannot finish the installation.", @"Quit", nil, nil);
		[NSApp terminate:self];
	}
	else if (err != noErr)
	{
		NSRunCriticalAlertPanel(nil, @"The installation could not complete because of issues encountered with the destination (permissions on certain files most likely).", @"Quit", nil, nil);
		[NSApp terminate:self];
	}
	else
	{
		NSRunInformationalAlertPanel(nil, @"Installation was successful.", nil, nil, nil);
	}
	
	[progBar stopAnimation:self];
	[pathField setEnabled:YES];
	[installButton setEnabled:NO];
	[locationButton setEnabled:YES];
}

@end
